import io
import json
import time
from typing import List, Optional

import pandas as pd

from mindfoundry.client.horizon.generated import Client
from mindfoundry.client.horizon.generated.api.pipelines import (
    delete_pipeline_resource,
    get_feature_export_resource,
    get_feature_nodes_and_links_resource,
    get_pipeline_resource,
    post_on_demand_predict_resource,
    post_predict_resource,
    post_restart_pipeline_resource,
    post_run_pipeline_resource,
    put_all_pipeline_resource,
    put_pipeline_lock_resource,
    put_stage_creation_resource,
)
from mindfoundry.client.horizon.generated.models import (
    BlueprintType,
    BoolLocked,
    ExpertPredictForm,
    NodeAndLink,
    OnDemandPredictForm,
    Pipeline,
    PipelineSpecification,
    PredictResponse,
    RegressorType,
    StageSpecification,
    StageStatus,
)

from ..generated.types import File
from .utils import assert_success, return_value_or_raise_error


class PipelinesClient:
    def __init__(self, client: Client):
        self._client = client

    def get(self, pipeline_id: int) -> Pipeline:
        pipeline = get_pipeline_resource.sync_detailed(
            client=self._client,
            id_=pipeline_id,
        )
        return return_value_or_raise_error(pipeline)

    def delete(self, pipeline_id: int) -> None:
        response = delete_pipeline_resource.sync_detailed(
            client=self._client,
            id_=pipeline_id,
        )
        assert_success(response)

    def create_custom(self, dataset_id: int, name: str) -> Pipeline:
        pipeline = put_all_pipeline_resource.sync_detailed(
            client=self._client,
            json_body=PipelineSpecification(
                name=name,
                dataset_id=dataset_id,
                blueprint=BlueprintType.CUSTOM,
            ),
        )
        return return_value_or_raise_error(pipeline)

    def add_stage(self, pipeline_id: int, spec: StageSpecification) -> Pipeline:
        pipeline = put_stage_creation_resource.sync_detailed(
            client=self._client,
            id_=pipeline_id,
            json_body=spec,
        )
        return return_value_or_raise_error(pipeline)

    def run(self, pipeline_id: int, wait_for_completion: bool = False) -> None:
        post_run_pipeline_resource.sync_detailed(client=self._client, id_=pipeline_id)
        if wait_for_completion:
            self._wait_for_completion(pipeline_id)

    def restart(self, pipeline_id: int, wait_for_completion: bool = False) -> None:
        post_restart_pipeline_resource.sync_detailed(client=self._client, id_=pipeline_id)
        if wait_for_completion:
            self._wait_for_completion(pipeline_id)

    def lock(self, pipeline_id: int) -> None:
        response = put_pipeline_lock_resource.sync_detailed(
            client=self._client,
            id_=pipeline_id,
            json_body=BoolLocked(locked=True),
        )
        assert_success(response)

    def predict(
        self,
        pipeline_id: int,
        horizon: int,
        regressor: RegressorType,
        new_data: Optional[pd.DataFrame] = None,
        filter_target_names: Optional[List[str]] = None,
    ) -> List[PredictResponse]:
        filter_target_names_as_json = (
            json.dumps(filter_target_names) if filter_target_names is not None else None
        )

        if new_data is None:
            prediction_response_list = post_on_demand_predict_resource.sync_detailed(
                client=self._client,
                id_=pipeline_id,
                json_body=OnDemandPredictForm(
                    horizon=horizon,
                    regressor=regressor,
                    filter_target_names=filter_target_names_as_json,
                ),
            )

            return return_value_or_raise_error(prediction_response_list).predict_responses
        else:
            str_buffer = io.StringIO(new_data.to_csv(encoding="utf-8", index=False))
            str_buffer.seek(0)
            str_buffer.name = "new data.csv"

            prediction_responses = post_predict_resource.sync_detailed(
                client=self._client,
                id_=pipeline_id,
                multipart_data=ExpertPredictForm(
                    horizon=horizon,
                    regressor=regressor,
                    filter_target_names=filter_target_names_as_json,
                    options="{}",
                    file=File(
                        payload=str_buffer,
                        file_name=str_buffer.name,
                        mime_type="text/csv",
                    ),
                ),
            )

            return return_value_or_raise_error(prediction_responses)

    def download_features(
        self,
        pipeline_id: int,
        horizon: int,
        include_targets: bool,
    ) -> pd.DataFrame:
        pipeline = self.get(pipeline_id)
        last_stage = pipeline.stages[-1]

        features_response = get_feature_export_resource.sync_detailed(
            client=self._client,
            id_=pipeline_id,
            stage_id=last_stage.id_,
            horizon=horizon,
            include_targets=include_targets,
        )

        data_file = return_value_or_raise_error(features_response)
        return pd.read_csv(data_file.payload)

    def get_feature_graph(self, pipeline_id: int) -> List[NodeAndLink]:
        pipeline = self.get(pipeline_id)
        last_stage = pipeline.stages[-1]

        graph_response = get_feature_nodes_and_links_resource.sync_detailed(
            client=self._client,
            id_=pipeline_id,
            stage_id=last_stage.id_,
        )

        graph_data = return_value_or_raise_error(graph_response)
        return graph_data.nodes_and_links

    def _wait_for_completion(self, pipeline_id: int) -> None:
        pipeline = self.get(pipeline_id)

        while not all(stage.status == StageStatus.COMPLETE for stage in pipeline.stages):
            time.sleep(2)

            pipeline = self.get(pipeline_id)
            errored_stages = [
                stage for stage in pipeline.stages if stage.status == StageStatus.ERROR
            ]

            if len(errored_stages) > 0:
                raise RuntimeError(errored_stages[0].error_msg)



from enum import Enum


class ClassificationScoreMetric(str, Enum):
    PRECISION_AT_20 = "precision_at_20"
    ACCURACY = "accuracy"
    BALANCED_ACCURACY_SCORE = "balanced_accuracy_score"
    F1_SCORE = "f1_score"
    PRECISION_SCORE = "precision_score"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.algo_execution_data import AlgoExecutionData
from ..types import UNSET, Unset

T = TypeVar("T", bound="AlgoExecutionTestCase")


@attr.s(auto_attribs=True)
class AlgoExecutionTestCase:
    """A basic set of inputs (mock predictions, mock history), which
    can be fed to an Algo to see how it will respond when called by the trading
    simulator."""

    input_id: int
    data: AlgoExecutionData
    name: Union[Unset, str] = ""
    description: Union[Unset, str] = ""
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        input_id = self.input_id
        data = self.data.to_dict()

        name = self.name
        description = self.description

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "inputId": input_id,
                "data": data,
            }
        )
        if name is not UNSET:
            field_dict["name"] = name
        if description is not UNSET:
            field_dict["description"] = description

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        input_id = d.pop("inputId")

        data = AlgoExecutionData.from_dict(d.pop("data"))

        name = d.pop("name", UNSET)

        description = d.pop("description", UNSET)

        algo_execution_test_case = cls(
            input_id=input_id,
            data=data,
            name=name,
            description=description,
        )

        algo_execution_test_case.additional_properties = d
        return algo_execution_test_case

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import Any, Dict, List, Type, TypeVar, Union

import attr
from dateutil.parser import isoparse

from ..models.ingestion_status import IngestionStatus
from ..types import UNSET, Unset

T = TypeVar("T", bound="IngestionProcess")


@attr.s(auto_attribs=True)
class IngestionProcess:
    """ An asynchronous process which parses, validates and profiles datasets on upload """

    id_: int
    status: IngestionStatus
    upload_size_bytes: int
    dataset_name: str
    creation_user_id: int
    creation_date: datetime.datetime
    last_update_error: Union[Unset, None, str] = UNSET
    error: Union[Unset, None, str] = UNSET
    task_id: Union[Unset, None, str] = UNSET
    dataset_id: Union[Unset, None, int] = UNSET
    last_update_date: Union[Unset, None, datetime.datetime] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id_ = self.id_
        status = self.status.value

        upload_size_bytes = self.upload_size_bytes
        dataset_name = self.dataset_name
        creation_user_id = self.creation_user_id
        creation_date = self.creation_date.isoformat()

        last_update_error = self.last_update_error
        error = self.error
        task_id = self.task_id
        dataset_id = self.dataset_id
        last_update_date: Union[Unset, None, str] = UNSET
        if not isinstance(self.last_update_date, Unset):
            last_update_date = self.last_update_date.isoformat() if self.last_update_date else None

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id_,
                "status": status,
                "uploadSizeBytes": upload_size_bytes,
                "datasetName": dataset_name,
                "creationUserId": creation_user_id,
                "creationDate": creation_date,
            }
        )
        if last_update_error is not UNSET:
            field_dict["lastUpdateError"] = last_update_error
        if error is not UNSET:
            field_dict["error"] = error
        if task_id is not UNSET:
            field_dict["taskId"] = task_id
        if dataset_id is not UNSET:
            field_dict["datasetId"] = dataset_id
        if last_update_date is not UNSET:
            field_dict["lastUpdateDate"] = last_update_date

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id_ = d.pop("id")

        status = IngestionStatus(d.pop("status"))

        upload_size_bytes = d.pop("uploadSizeBytes")

        dataset_name = d.pop("datasetName")

        creation_user_id = d.pop("creationUserId")

        creation_date = isoparse(d.pop("creationDate"))

        last_update_error = d.pop("lastUpdateError", UNSET)

        error = d.pop("error", UNSET)

        task_id = d.pop("taskId", UNSET)

        dataset_id = d.pop("datasetId", UNSET)

        last_update_date = None
        _last_update_date = d.pop("lastUpdateDate", UNSET)
        if _last_update_date is not None and not isinstance(_last_update_date, Unset):
            last_update_date = isoparse(_last_update_date)

        ingestion_process = cls(
            id_=id_,
            status=status,
            upload_size_bytes=upload_size_bytes,
            dataset_name=dataset_name,
            creation_user_id=creation_user_id,
            creation_date=creation_date,
            last_update_error=last_update_error,
            error=error,
            task_id=task_id,
            dataset_id=dataset_id,
            last_update_date=last_update_date,
        )

        ingestion_process.additional_properties = d
        return ingestion_process

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from mindfoundry.client.horizon.generated.models import (
    ClassificationDiscoveryConfig as GeneratedClassificationDiscoveryConfig,
)
from mindfoundry.client.horizon.generated.models import (
    CorrelationMethod as GeneratedCorrelationMethod,
)
from mindfoundry.client.horizon.generated.models import (
    FeatureGenerationStageConfig as GeneratedFeatureGenerationStageConfig,
)
from mindfoundry.client.horizon.generated.models import (
    FeatureGeneratorType as GeneratedFeatureGeneratorType,
)
from mindfoundry.client.horizon.generated.models import (
    FilterStageConfig as GeneratedFilterStageConfig,
)
from mindfoundry.client.horizon.generated.models import (
    RefinementStageConfig as GeneratedRefinementStageConfig,
)
from mindfoundry.client.horizon.generated.models import RegressorType as GeneratedRegressorType
from mindfoundry.client.horizon.generated.models import (
    StationarisationStageConfig as GeneratedStationarisationStageConfig,
)
from mindfoundry.client.horizon.generated.models import (
    StationarisationStrategy as GeneratedStationarisationStrategy,
)
from mindfoundry.client.horizon.generated.models import (
    TargetTransformType as GeneratedTargetTransformType,
)

from .base import ValidatedConfig, assert_not_empty_if_provided
from .values import (
    CorrelationMethod,
    FeatureGeneratorType,
    RegressorType,
    StationarizationStrategy,
    TargetTransformType,
)


@dataclass
class StationarizationConfig(ValidatedConfig):
    """ Configure the stationarization stage of a forecasting model. """

    strategy: Optional[StationarizationStrategy] = None
    """
    What strategy should be used to handle columns that support / do not support stationarization?
    Default is `keep_fail`.
    """

    target_transform: Optional[TargetTransformType] = None
    """
    Whether the target should undergo a lag transformation as well. Default is `HorizonLagDiff`.
    """

    adf_threshold: Optional[float] = None
    """
    Running stationarisation of features runs an ADF test on all the features. An ADF test
    is a statistical test that assigns a score between 0 and 1 to each feature, where scores
    closer to zero are more likely stationary, and scores closer to 1 are more likely non
    stationary. The score itself is a statistical p-value, and this threshold specifies
    the p-value above which a feature is considered non stationary.
    Default is 0.03.
    """

    def validate(self) -> None:
        if self.adf_threshold is not None:
            assert (
                0 <= self.adf_threshold <= 1
            ), "The adf_threshold of the stationarization stage must be between 0 and 1"

    def to_stage_config(self) -> GeneratedStationarisationStageConfig:
        default_stage_config = self.default_stage_config()
        return GeneratedStationarisationStageConfig(
            adf_threshold=self.adf_threshold or default_stage_config.adf_threshold,
            strategy=(
                GeneratedStationarisationStrategy(self.strategy)
                if self.strategy is not None
                else default_stage_config.strategy
            ),
            target_transform=(
                GeneratedTargetTransformType(self.target_transform)
                if self.target_transform is not None
                else default_stage_config.target_transform
            ),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "strategy": self.strategy,
            "target_transform": self.target_transform,
            "adf_threshold": self.adf_threshold,
        }

    @staticmethod
    def from_dict(dictionary: Dict[str, Any]) -> StationarizationConfig:
        return StationarizationConfig(
            strategy=dictionary["strategy"],
            target_transform=dictionary["target_transform"],
            adf_threshold=dictionary["adf_threshold"],
        )

    @staticmethod
    def default() -> StationarizationConfig:
        return StationarizationConfig(
            adf_threshold=0.03,
            strategy=StationarizationStrategy.KEEP_FAILED,
            target_transform=TargetTransformType.HORIZON_LAG_DIFF,
        )

    @staticmethod
    def default_stage_config() -> GeneratedStationarisationStageConfig:
        return GeneratedStationarisationStageConfig(
            adf_threshold=0.03,
            strategy=GeneratedStationarisationStrategy.KEEP_FAIL,
            target_transform=GeneratedTargetTransformType.HORIZONLAGDIFF,
        )


@dataclass
class FeatureGenerationConfig(ValidatedConfig):
    """ Configure the feature generation stage of a forecasting model. """

    max_features: Optional[int] = None
    """ Upper bound on the number of features to generate per horizon. Default is 500. """

    generators: Optional[List[FeatureGeneratorType]] = None
    """
    List of all feature generation strategies to be used in this stage. By default, all
    generators will be used.
    """

    def validate(self) -> None:
        if self.max_features is not None:
            assert self.max_features > 0, (
                "The max_features parameter of the feature generation stage must be "
                "greater than zero"
            )

        if self.generators is not None:
            assert len(self.generators) > 0, (
                "When specifying generators used in the feature generation stage, at least "
                "one generator must be provided"
            )

            # fmt: off
            assert len(set(self.generators)) == len(self.generators), (
                "The list of generators should not contain any duplicate item."
            )
            # fmt: on

    def to_stage_config(self) -> GeneratedFeatureGenerationStageConfig:
        default_stage_config = self.default_stage_config()
        return GeneratedFeatureGenerationStageConfig(
            max_n_features=self.max_features or default_stage_config.max_n_features,
            feature_generators=(
                [GeneratedFeatureGeneratorType(g) for g in self.generators]
                if self.generators is not None
                else default_stage_config.feature_generators
            ),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "max_features": self.max_features,
            "generators": self.generators,
        }

    @staticmethod
    def from_dict(dictionary: Dict[str, Any]) -> FeatureGenerationConfig:
        return FeatureGenerationConfig(
            max_features=dictionary["max_features"],
            generators=dictionary["generators"],
        )

    @staticmethod
    def default() -> FeatureGenerationConfig:
        return FeatureGenerationConfig(
            max_features=500,
            generators=[
                FeatureGeneratorType.LAG,
                FeatureGeneratorType.EWMA,
                FeatureGeneratorType.AUTO_LAG,
                FeatureGeneratorType.CALENDAR,
                FeatureGeneratorType.LOGARITHM,
                FeatureGeneratorType.NUMBER_OF_PEAKS,
                FeatureGeneratorType.ONE_HOT_ENCODING,
                FeatureGeneratorType.PERCENT_CHANGE,
                FeatureGeneratorType.ROLLING_AVERAGE,
            ],
        )

    @staticmethod
    def default_stage_config() -> GeneratedFeatureGenerationStageConfig:
        return GeneratedFeatureGenerationStageConfig(
            max_n_features=500,
            feature_generators=[
                GeneratedFeatureGeneratorType.LAG,
                GeneratedFeatureGeneratorType.EWMA,
                GeneratedFeatureGeneratorType.AUTOLAG,
                GeneratedFeatureGeneratorType.CALENDAR,
                GeneratedFeatureGeneratorType.LOGARITHM,
                GeneratedFeatureGeneratorType.NUM_PEAKS,
                GeneratedFeatureGeneratorType.ONE_HOT_ENCODE,
                GeneratedFeatureGeneratorType.PERC_CHANGE,
                GeneratedFeatureGeneratorType.ROLLING_AVERAGE,
            ],
        )


@dataclass
class FilteringConfig(ValidatedConfig):
    """ Configure the filtering stage of a forecasting model. """

    max_features: Optional[int] = None
    """ Upper bound on the number of remaining features per horizon. Default is 30. """

    correlation_method: Optional[CorrelationMethod] = None
    """
    Correlation method to be used to determine the most appropriate features to be filtered out
    and the most useful features to keep. Default is `pearson`.
    """

    def validate(self) -> None:
        if self.max_features is not None:
            assert (
                self.max_features > 0
            ), "The max_features parameter of the filtering stage must be greater than zero"

    def to_stage_config(self) -> GeneratedFilterStageConfig:
        default_stage_config = self.default_stage_config()
        return GeneratedFilterStageConfig(
            max_n_features=self.max_features or default_stage_config.max_n_features,
            method=(
                GeneratedCorrelationMethod(self.correlation_method)
                if self.correlation_method is not None
                else default_stage_config.method
            ),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "max_features": self.max_features,
            "correlation_method": self.correlation_method,
        }

    @staticmethod
    def from_dict(dictionary: Dict[str, Any]) -> FilteringConfig:
        return FilteringConfig(
            max_features=dictionary["max_features"],
            correlation_method=dictionary["correlation_method"],
        )

    @staticmethod
    def default() -> FilteringConfig:
        return FilteringConfig(
            max_features=30,
            correlation_method=CorrelationMethod.PEARSON,
        )

    @staticmethod
    def default_stage_config() -> GeneratedFilterStageConfig:
        return GeneratedFilterStageConfig(
            max_n_features=30,
            method=GeneratedCorrelationMethod.PEARSON,
        )


@dataclass
class RefinementConfig(ValidatedConfig):
    """ Configure the refinement stage of a forecasting model. """

    max_features: Optional[int] = None
    """ Upper bound on number of remaining features per horizon. Default is 20. """

    min_features: Optional[int] = None
    """ Lower bound on the number of remaining features per horizon. Default is 5. """

    early_stopping_sensitivity: Optional[float] = None
    """
    This must be a value between 0 and 1. If closer to 1 then the stage will run faster,
    but will result in a larger feature set. Default is `0`.
    """

    deep_search: Optional[bool] = None
    """
    Whether to try dropping each feature at every iteration, or only a subset (50%).
    Default is `false`.
    """

    regressor: Optional[RegressorType] = None
    """ Regressor to be used for the refinement. Default is `RandomForest`. """

    def validate(self) -> None:
        if self.early_stopping_sensitivity is not None:
            assert 0 <= self.early_stopping_sensitivity <= 1, (
                "The early_stopping_sensitivity parameter of the refinement stage must be "
                "between 0 and 1 (bounds included)"
            )

        if self.min_features is not None:
            assert (
                self.min_features > 0
            ), "The min_features parameter of the refinement stage must be greater than zero"

        if self.max_features is not None:
            assert (
                self.max_features > 0
            ), "The max_features parameter of the refinement stage must be greater than zero"

        if self.max_features is not None and self.min_features is not None:
            assert 0 < self.min_features <= self.max_features, (
                "The min_features parameter of the refinement stage must be greater than "
                "zero and less than (or equal to) max_features"
            )

    def to_stage_config(self) -> GeneratedRefinementStageConfig:
        default_stage_config = self.default_stage_config()
        return GeneratedRefinementStageConfig(
            max_features=self.max_features or default_stage_config.max_features,
            min_features=self.min_features or default_stage_config.min_features,
            early_stopping_sensitivity=(
                self.early_stopping_sensitivity
                if self.early_stopping_sensitivity is not None
                else default_stage_config.early_stopping_sensitivity
            ),
            deep_search=(
                self.deep_search
                if self.deep_search is not None
                else default_stage_config.deep_search
            ),
            regressor=(
                GeneratedRegressorType(self.regressor)
                if self.regressor is not None
                else default_stage_config.regressor
            ),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "max_features": self.max_features,
            "min_features": self.min_features,
            "early_stopping_sensitivity": self.early_stopping_sensitivity,
            "deep_search": self.deep_search,
            "regressor": self.regressor,
        }

    @staticmethod
    def from_dict(dictionary: Dict[str, Any]) -> RefinementConfig:
        return RefinementConfig(
            max_features=dictionary["max_features"],
            min_features=dictionary["min_features"],
            early_stopping_sensitivity=dictionary["early_stopping_sensitivity"],
            deep_search=dictionary["deep_search"],
            regressor=dictionary["regressor"],
        )

    @staticmethod
    def default() -> RefinementConfig:
        return RefinementConfig(
            max_features=20,
            min_features=5,
            early_stopping_sensitivity=0,
            deep_search=False,
            regressor=RegressorType.RANDOM_FOREST,
        )

    @staticmethod
    def default_stage_config() -> GeneratedRefinementStageConfig:
        return GeneratedRefinementStageConfig(
            max_features=20,
            min_features=5,
            early_stopping_sensitivity=0,
            deep_search=False,
            regressor=GeneratedRegressorType.RANDOMFOREST,
        )


@dataclass
class ClassificationDiscoveryConfig(ValidatedConfig):
    """ Configure the discovery stage of a classification model. """

    max_features: Optional[int] = None
    """ Upper bound on the number of features to generate per horizon. Default is 100. """

    timeout_seconds: Optional[float] = None
    """ Maximum time allowed for running this stage. Default is `60` seconds. """

    excluded_columns: Optional[List[str]] = None
    """
    List of column names to exclude from feature generation.
    By default, no column will be excluded.
    """

    feature_generation_enabled: Optional[bool] = None
    """
    Discovery can run either with or without feature generation. In the latter case, only the
    original columns will be taken into account and Horizon will not generate any artificial
    feature. Default is True.
    """

    def validate(self) -> None:
        assert_not_empty_if_provided("excluded_columns", self.excluded_columns)

        if self.timeout_seconds is not None:
            assert self.timeout_seconds > 0, "Timeout must be a positive value"

    def to_stage_config(self) -> GeneratedClassificationDiscoveryConfig:
        default_stage_config = self.default_stage_config()
        return GeneratedClassificationDiscoveryConfig(
            max_n_features=self.max_features or default_stage_config.max_n_features,
            timeout_seconds=self.timeout_seconds or default_stage_config.timeout_seconds,
            columns_to_not_generate_features_from=(
                self.excluded_columns or default_stage_config.columns_to_not_generate_features_from
            ),
            feature_generation_enabled=(
                self.feature_generation_enabled
                if self.feature_generation_enabled is not None
                else default_stage_config.feature_generation_enabled
            ),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "max_features": self.max_features,
            "timeout_seconds": self.timeout_seconds,
            "excluded_columns": self.excluded_columns,
            "feature_generation_enabled": self.feature_generation_enabled,
        }

    @staticmethod
    def from_dict(dictionary: Dict[str, Any]) -> ClassificationDiscoveryConfig:
        return ClassificationDiscoveryConfig(
            max_features=dictionary["max_features"],
            timeout_seconds=dictionary["timeout_seconds"],
            excluded_columns=dictionary["excluded_columns"],
            feature_generation_enabled=dictionary["feature_generation_enabled"],
        )

    @staticmethod
    def default() -> ClassificationDiscoveryConfig:
        return ClassificationDiscoveryConfig(
            max_features=100,
            timeout_seconds=60,
            excluded_columns=[],
            feature_generation_enabled=True,
        )

    @staticmethod
    def default_stage_config() -> GeneratedClassificationDiscoveryConfig:
        return GeneratedClassificationDiscoveryConfig(
            timeout_seconds=60,
            feature_generation_enabled=True,
            columns_to_not_generate_features_from=[],
            max_n_features=100,
        )



from enum import Enum


class TradingAlgoParamType(str, Enum):
    STR_ = "str_"
    INT_ = "int_"
    FLOAT_ = "float_"
    BOOL_ = "bool_"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, Optional

import httpx

from ...client import Client
from ...models.dataset import Dataset
from ...types import UNSET, Response


def _get_kwargs(
    *,
    client: Client,
    id_: int,
    new_name: str,
) -> Dict[str, Any]:
    url = "{}/datasets/{id}/name".format(client.base_url, id=id_)

    headers: Dict[str, Any] = client.get_headers()
    cookies: Dict[str, Any] = client.get_cookies()

    params: Dict[str, Any] = {
        "newName": new_name,
    }
    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    return {
        "url": url,
        "headers": headers,
        "cookies": cookies,
        "timeout": client.get_timeout(),
        "params": params,
    }


def _parse_response(*, response: httpx.Response) -> Optional[Dataset]:
    if response.status_code == 200:
        response_200 = Dataset.from_dict(response.json())

        return response_200
    return None


def _build_response(*, response: httpx.Response) -> Response[Dataset]:
    return Response(
        status_code=response.status_code,
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(response=response),
    )


def sync_detailed(
    *,
    client: Client,
    id_: int,
    new_name: str,
) -> Response[Dataset]:
    kwargs = _get_kwargs(
        client=client,
        id_=id_,
        new_name=new_name,
    )

    response = httpx.put(
        **kwargs,
    )

    return _build_response(response=response)


def sync(
    *,
    client: Client,
    id_: int,
    new_name: str,
) -> Optional[Dataset]:
    """  """

    return sync_detailed(
        client=client,
        id_=id_,
        new_name=new_name,
    ).parsed


async def asyncio_detailed(
    *,
    client: Client,
    id_: int,
    new_name: str,
) -> Response[Dataset]:
    kwargs = _get_kwargs(
        client=client,
        id_=id_,
        new_name=new_name,
    )

    async with httpx.AsyncClient() as _client:
        response = await _client.put(**kwargs)

    return _build_response(response=response)


async def asyncio(
    *,
    client: Client,
    id_: int,
    new_name: str,
) -> Optional[Dataset]:
    """  """

    return (
        await asyncio_detailed(
            client=client,
            id_=id_,
            new_name=new_name,
        )
    ).parsed



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.classification_backtest_config import ClassificationBacktestConfig
from ..models.stage_status import StageStatus
from ..types import UNSET, Unset

T = TypeVar("T", bound="ClassificationBacktestStage")


@attr.s(auto_attribs=True)
class ClassificationBacktestStage:
    """Stage for testing the classifier over a range of time

    For each timestamp within the date range provided, it will:
    - Use all rows in its past to train a model to predict the given target
    - Predict the target for all the rows whose index is that timestamp

    Finally all results are concatenated (in timestamp order)

    It creates insights including all prediction probabilities, as well as some rolling
    and summary metrics.

    All results (predictions, ground_truth, confusion matrix) are to be interpreted as
        * true = value equals the `labelToPredict`
        * false = value does not equal `labelToPredict`"""

    status: StageStatus
    id_: int
    config: ClassificationBacktestConfig
    error_msg: Union[Unset, None, str] = UNSET
    type_: Union[Unset, str] = "classification_backtest"
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        status = self.status.value

        id_ = self.id_
        config = self.config.to_dict()

        error_msg = self.error_msg
        type_ = self.type_

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "status": status,
                "id": id_,
                "config": config,
            }
        )
        if error_msg is not UNSET:
            field_dict["errorMsg"] = error_msg
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        status = StageStatus(d.pop("status"))

        id_ = d.pop("id")

        config = ClassificationBacktestConfig.from_dict(d.pop("config"))

        error_msg = d.pop("errorMsg", UNSET)

        type_ = d.pop("type", UNSET)

        classification_backtest_stage = cls(
            status=status,
            id_=id_,
            config=config,
            error_msg=error_msg,
            type_=type_,
        )

        classification_backtest_stage.additional_properties = d
        return classification_backtest_stage

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.regressor_type import RegressorType
from ..types import UNSET, Unset

T = TypeVar("T", bound="RefinementStageConfig")


@attr.s(auto_attribs=True)
class RefinementStageConfig:
    """ Properties for a stage  """

    min_features: int
    max_features: int
    early_stopping_sensitivity: float
    deep_search: bool
    regressor: RegressorType
    type_: Union[Unset, str] = "refinement"
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        min_features = self.min_features
        max_features = self.max_features
        early_stopping_sensitivity = self.early_stopping_sensitivity
        deep_search = self.deep_search
        regressor = self.regressor.value

        type_ = self.type_

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "minFeatures": min_features,
                "maxFeatures": max_features,
                "earlyStoppingSensitivity": early_stopping_sensitivity,
                "deepSearch": deep_search,
                "regressor": regressor,
            }
        )
        if type_ is not UNSET:
            field_dict["type"] = type_

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        min_features = d.pop("minFeatures")

        max_features = d.pop("maxFeatures")

        early_stopping_sensitivity = d.pop("earlyStoppingSensitivity")

        deep_search = d.pop("deepSearch")

        regressor = RegressorType(d.pop("regressor"))

        type_ = d.pop("type", UNSET)

        refinement_stage_config = cls(
            min_features=min_features,
            max_features=max_features,
            early_stopping_sensitivity=early_stopping_sensitivity,
            deep_search=deep_search,
            regressor=regressor,
            type_=type_,
        )

        refinement_stage_config.additional_properties = d
        return refinement_stage_config

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import json
from typing import Optional, TypeVar

from mindfoundry.client.horizon.generated.types import Response

SomeType = TypeVar("SomeType")


class ApiError(RuntimeError):
    def __init__(self, message: str, ref: str):
        super().__init__(message)
        self._ref = ref

    @property
    def ref(self) -> str:
        return self._ref


def non_optional(value: Optional[SomeType]) -> SomeType:
    assert value is not None
    return value


def assert_success(response: Response[SomeType]) -> None:
    if response.status_code < 200 or response.status_code >= 300:
        content = json.loads(response.content)
        if "summary" in content and "ref" in content:
            raise ApiError(content["summary"], content["ref"])
        else:
            raise RuntimeError(f"Unknown api error. Status code {response.status_code}")


def return_value_or_raise_error(response: Response[SomeType]) -> SomeType:
    assert_success(response)
    return non_optional(response.parsed)



from typing import Any, Dict, List, Type, TypeVar, Union

import attr

from ..models.json_data_type import JsonDataType
from ..types import UNSET, Unset

T = TypeVar("T", bound="JsonDataDtype")


@attr.s(auto_attribs=True)
class JsonDataDtype:
    """  """

    name: str
    type_: JsonDataType
    date_format: Union[Unset, None, str] = UNSET
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        type_ = self.type_.value

        date_format = self.date_format

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "type": type_,
            }
        )
        if date_format is not UNSET:
            field_dict["dateFormat"] = date_format

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        type_ = JsonDataType(d.pop("type"))

        date_format = d.pop("dateFormat", UNSET)

        json_data_dtype = cls(
            name=name,
            type_=type_,
            date_format=date_format,
        )

        json_data_dtype.additional_properties = d
        return json_data_dtype

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import Any, Dict, List, Type, TypeVar

import attr
from dateutil.parser import isoparse

from ..models.blueprint_type import BlueprintType

T = TypeVar("T", bound="PipelineSummary")


@attr.s(auto_attribs=True)
class PipelineSummary:
    """ High level information about a pipeline  """

    id_: int
    name: str
    locked: bool
    auto_update: bool
    blueprint: BlueprintType
    creation_date: datetime.datetime
    creation_user: int
    dataset_name: str
    additional_properties: Dict[str, Any] = attr.ib(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id_ = self.id_
        name = self.name
        locked = self.locked
        auto_update = self.auto_update
        blueprint = self.blueprint.value

        creation_date = self.creation_date.isoformat()

        creation_user = self.creation_user
        dataset_name = self.dataset_name

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id_,
                "name": name,
                "locked": locked,
                "autoUpdate": auto_update,
                "blueprint": blueprint,
                "creationDate": creation_date,
                "creationUser": creation_user,
                "datasetName": dataset_name,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id_ = d.pop("id")

        name = d.pop("name")

        locked = d.pop("locked")

        auto_update = d.pop("autoUpdate")

        blueprint = BlueprintType(d.pop("blueprint"))

        creation_date = isoparse(d.pop("creationDate"))

        creation_user = d.pop("creationUser")

        dataset_name = d.pop("datasetName")

        pipeline_summary = cls(
            id_=id_,
            name=name,
            locked=locked,
            auto_update=auto_update,
            blueprint=blueprint,
            creation_date=creation_date,
            creation_user=creation_user,
            dataset_name=dataset_name,
        )

        pipeline_summary.additional_properties = d
        return pipeline_summary

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class StageType(str, Enum):
    FEATURE_GENERATION = "feature_generation"
    FILTERING = "filtering"
    PROBLEM_SPECIFICATION = "problem_specification"
    STATIONARISATION = "stationarisation"
    BACKTEST = "backtest"
    LSTM_BACKTEST = "lstm_backtest"
    REFINEMENT = "refinement"
    PREDICTION = "prediction"
    LSTM_PREDICTION = "lstm_prediction"
    TRADING_SPECIFICATION = "trading_specification"
    TRADING_SIMULATION = "trading_simulation"
    CLASSIFICATION_SPECIFICATION = "classification_specification"
    CLASSIFICATION_DISCOVERY = "classification_discovery"
    CLASSIFICATION_BACKTEST = "classification_backtest"

    def __str__(self) -> str:
        return str(self.value)



from enum import Enum


class TargetTransformType(str, Enum):
    DONOTHING = "DoNothing"
    HORIZONLAGDIFF = "HorizonLagDiff"
    HORIZONLAGDIFFRATIO = "HorizonLagDiffRatio"

    def __str__(self) -> str:
        return str(self.value)
